"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const electron_1 = require("electron");
const log_1 = require("./log");
function partitionForAccount(account) {
    let host = account.host;
    if (account.host.startsWith('https://')) {
        host = host.slice(8);
    }
    else if (account.host.startsWith('http://')) {
        host = host.slice(7);
    }
    return `persist:mstdn:${account.name}:${host}`;
}
exports.partitionForAccount = partitionForAccount;
class AccountSwitcher extends events_1.EventEmitter {
    constructor(accounts) {
        super();
        this.switchToNext = () => {
            if (this.accounts.length <= 1) {
                log_1.default.debug('Skip switching account: Only one account is registered');
                return;
            }
            let idx = this.getAccountIndex(this.current) + 1;
            if (idx >= this.accounts.length) {
                idx = 0;
            }
            log_1.default.debug('Switch to next account');
            this.switchTo(this.accounts[idx]);
        };
        this.switchToPrev = () => {
            if (this.accounts.length <= 1) {
                log_1.default.debug('Skip switching account: Only one account is registered');
                return;
            }
            let idx = this.getAccountIndex(this.current);
            if (idx <= 0) {
                idx = this.accounts.length;
            }
            --idx;
            log_1.default.debug('Switch to prev account');
            this.switchTo(this.accounts[idx]);
        };
        const submenu = [];
        for (const account of accounts.filter(a => a.name !== '')) {
            let name = account.name;
            if (!name.startsWith('@')) {
                name = '@' + name;
            }
            name += '@' + account.host;
            submenu.push({
                label: name,
                type: 'radio',
                checked: false,
                click: () => this.switchTo(account),
            });
        }
        this.accounts = accounts;
        if (accounts.length === 0) {
            return;
        }
        submenu[0].checked = true;
        this.current = accounts[0];
        const item = new electron_1.MenuItem({
            label: 'Accounts',
            type: 'submenu',
            submenu,
        });
        const menu = electron_1.Menu.getApplicationMenu();
        if (menu !== null) {
            menu.insert(menu.items.length - 1, item);
            electron_1.Menu.setApplicationMenu(menu);
        }
        electron_1.ipcMain.on('mstdn:next-account', this.switchToNext);
        electron_1.ipcMain.on('mstdn:prev-account', this.switchToPrev);
    }
    switchTo(account) {
        if (this.current.name === account.name && this.current.host === account.host) {
            log_1.default.debug('Current account is already @' + account.name);
            return;
        }
        log_1.default.debug('Switch to account', account);
        this.emit('switch', account, this.current);
        this.current = account;
    }
    getAccountIndex(account) {
        for (let i = 0; i < this.accounts.length; ++i) {
            const a = this.accounts[i];
            if (a.name === account.name && a.host === account.host) {
                return i;
            }
        }
        return -1;
    }
}
exports.default = AccountSwitcher;
//# sourceMappingURL=account_switcher.js.map